import datetime
import json
import os

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import GObject, GLib


class TrackingState(GObject.GObject):
    """The state of the time tracking.

    Includes the start time, the task and the client.
    Whenever the state changes, the state is saved to a json file.
    When the start time is None, the time is not being tracked.
    """
    start_time = GObject.Property(type=GLib.DateTime, default=None)
    task = GObject.Property(type=str, default="")
    client = GObject.Property(type=str, default="")

    def __init__(self, path: str):
        GObject.GObject.__init__(self)
        self._filename = path
        self._load()
        self.connect("notify", self._save)

    def _load(self):
        if os.path.exists(self._filename):
            with open(self._filename, "r") as f:
                d = json.load(f)
                self.start_time = GLib.DateTime.new_from_iso8601(d["start_time"]) if d["start_time"] else None
                self.task = d["task"]
                self.client = d["client"]

    def _save(self, *_args):
        with open(self._filename, "w") as f:
            json.dump({
            "start_time": self.start_time.format_iso8601() if self.start_time else None,
            "task": self.task,
            "client": self.client,
        }, f, indent=2)

    def is_tracking(self) -> bool:
        return self.start_time is not None

    def is_client_and_task_set(self) -> bool:
        return self.client != "" and self.task != ""

    def elapsed_time(self) -> datetime.timedelta:
        if self.start_time is None:
            return datetime.timedelta()
        return datetime.timedelta(microseconds=GLib.DateTime.new_now_local().difference(self.start_time))



import datetime
import gi

from wage_labor_record.tracking_state import TrackingState
from wage_labor_record.utils import make_completer
from wage_labor_record.worked_time_store import WorkedTimeStore

gi.require_version("Gtk", "3.0")
from gi.repository import GLib, GObject, Gtk


class TimeTrackerWindow(Gtk.Dialog):
    def __init__(self, tracking_state: TrackingState, worked_time_store: WorkedTimeStore, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Layout
        box = self.get_content_area()

        self.task_entry = Gtk.Entry(
            placeholder_text="Task",
            completion=make_completer(worked_time_store.tasks),
        )
        self.task_entry.connect("activate", lambda *_args: self.start_tracking_button.clicked())
        self.task_entry.show()
        box.add(self.task_entry)

        self.client_entry = Gtk.Entry(
            placeholder_text="Client",
            completion=make_completer(worked_time_store.clients),
        )
        self.client_entry.connect("activate", lambda *_args: self.start_tracking_button.clicked())
        self.client_entry.show()
        box.add(self.client_entry)

        self.action_bar = Gtk.ActionBar()
        self.action_bar.show()
        box.add(self.action_bar)

        # Green Start Button
        self.start_tracking_button = Gtk.Button(label="Start")
        self.start_tracking_button.show()
        self.start_tracking_button.set_action_name("app.start_tracking")
        self.action_bar.pack_start(self.start_tracking_button)
        self.start_tracking_button.get_style_context().add_class("suggested-action")

        # Red Stop Button
        self.stop_tracking_button = Gtk.Button(label="Stop")
        self.stop_tracking_button.show()
        self.stop_tracking_button.set_action_name("app.stop_tracking")
        self.action_bar.pack_start(self.stop_tracking_button)
        self.stop_tracking_button.get_style_context().add_class("destructive-action")

        # Abort Button
        self.abort_tracking_button = Gtk.Button(label="Abort")
        self.abort_tracking_button.show()
        self.abort_tracking_button.set_action_name("app.abort_tracking")
        self.action_bar.pack_start(self.abort_tracking_button)

        # A Monospaced bold label with large font size
        self.elapsed_time_label = Gtk.Label(label="")
        self.elapsed_time_label.show()
        box.pack_start(self.elapsed_time_label, True, True, 0)

        # Ensure the entry fields edit the action properties
        tracking_state.bind_property("task", self.task_entry, "text", GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE)
        tracking_state.bind_property("client", self.client_entry, "text", GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE)

        # When the tracking is active, repeatedly update the elapsed time label
        def _update_elapsed_time_label():
            if tracking_state.is_tracking():
                label_txt = str(datetime.timedelta(
                    seconds=
                        round(GLib.DateTime.new_now_local().difference(tracking_state.start_time)/1000000)
                ))
                self.elapsed_time_label.set_markup(f"<span font='monospace bold 24'>{label_txt}</span>")
            return tracking_state.is_tracking()

        def _setup_elapsed_time_label_updates(*_):
            if tracking_state.is_tracking():
                _update_elapsed_time_label()
                GLib.timeout_add(1000, _update_elapsed_time_label)
            else:
                label_txt = str(datetime.timedelta(seconds=0))
                self.elapsed_time_label.set_markup(f"<span font='monospace bold 24'>{label_txt}</span>")

        _setup_elapsed_time_label_updates()
        tracking_state.connect("notify::start-time", _setup_elapsed_time_label_updates)



import logging

import gi

from wage_labor_record.worked_time_store import WorkedTime

gi.require_version("Gtk", "3.0")
gi.require_version('XApp', '1.0')
from gi.repository import Gio, GLib, GObject

from wage_labor_record.tracking_state import TrackingState


class AbortTrackingAction(Gio.SimpleAction):
    def __init__(self, tracking_state: TrackingState):
        super().__init__(name="abort_tracking", parameter_type=None, state=None)
        self._tracking_state = tracking_state
        self.connect("activate", self._abort_tracking)
        tracking_state.connect("notify::start-time", self._update_enabled_state)
        self._update_enabled_state()

    def _abort_tracking(self, *_args):
        self._tracking_state.start_time = None
        logging.debug("Abort tracking")

    def _update_enabled_state(self, *_args):
        self.set_enabled(self._tracking_state.is_tracking())
        if self.get_enabled():
            logging.debug("Abort tracking enabled")
        else:
            logging.debug("Abort tracking disabled")


class StopTrackingAction(Gio.SimpleAction):
    worked_time = GObject.Signal("worked-time", arg_types=[GObject.TYPE_PYOBJECT])

    def __init__(self, tracking_state: TrackingState):
        super().__init__(name="stop_tracking", parameter_type=None, state=None)
        self._tracking_state = tracking_state
        self.connect("activate", self._stop_tracking)
        tracking_state.connect("notify", self._update_enabled_state)
        self._update_enabled_state()

    def _stop_tracking(self, *_args):
        assert self.get_enabled()
        assert self._tracking_state.start_time is not None
        assert self._tracking_state.task != ""
        assert self._tracking_state.client != ""

        self.emit("worked-time", WorkedTime(
            self._tracking_state.task,
            self._tracking_state.client,
            self._tracking_state.start_time,
            GLib.DateTime.new_now_local(),
        ))

        self._tracking_state.start_time = None

        logging.debug("Stop tracking")

    def _update_enabled_state(self, *_args):
        self.set_enabled(self._tracking_state.is_client_and_task_set() and self._tracking_state.is_tracking())
        if self.get_enabled():
            logging.debug("Stop tracking enabled")
        else:
            logging.debug("Stop tracking disabled")


class SetCurrentTaskAction(Gio.SimpleAction):
    """
    Action to set the current task and start tracking it.
    Can only be used if the current task is not set, or we are currently not tracking any task.
    """
    def __init__(self, tracking_state: TrackingState):
        super().__init__(name="start_tracking_task", parameter_type=GLib.VariantType("(ss)"), state=None)
        self._tracking_state = tracking_state
        self.connect("activate", self._start_tracking_task)
        tracking_state.connect("notify", self._update_enabled_state)
        self._update_enabled_state()

    def _start_tracking_task(self, _action, parameter):
        client, task = parameter
        self._tracking_state.task = task
        self._tracking_state.client = client
        if not self._tracking_state.is_tracking():
            self._tracking_state.start_time = GLib.DateTime.new_now_local()
        logging.debug(f"Start tracking task {self._tracking_state.task}")

    def _update_enabled_state(self, *_args):
        self.set_enabled(not self._tracking_state.is_tracking() or not self._tracking_state.is_client_and_task_set())

        if self.get_enabled():
            logging.debug("Start tracking task enabled")
        else:
            logging.debug("Start tracking task disabled")


class StartTrackingAction(Gio.SimpleAction):
    def __init__(self, tracking_state: TrackingState):
        super().__init__(name="start_tracking", parameter_type=None, state=None)
        self._tracking_state = tracking_state
        self.connect("activate", self._start_tracking)
        tracking_state.connect("notify::start-time", self._update_enabled_state)
        self._update_enabled_state()

    def _start_tracking(self, *_args):
        self._tracking_state.start_time = GLib.DateTime.new_now_local()
        logging.debug("Start tracking")

    def _update_enabled_state(self, *_args):
        self.set_enabled(not self._tracking_state.is_tracking())
        if self.get_enabled():
            logging.debug("Start tracking enabled")
        else:
            logging.debug("Start tracking disabled")


import json
import logging
import os
from datetime import timedelta
from typing import Callable, Generator, Optional, Set, Tuple

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gio, GLib, GObject


class WorkedTime(GObject.GObject):
    task = GObject.Property(type=str, default="")
    client = GObject.Property(type=str, default="")
    start_time = GObject.Property(type=GLib.DateTime, default=None)
    end_time = GObject.Property(type=GLib.DateTime, default=None)

    def __init__(self,  task: str, client: str, start_time: GLib.DateTime, end_time: GLib.DateTime):
        GObject.GObject.__init__(self)
        self.task = task
        self.client = client
        self.start_time = start_time
        self.end_time = end_time

        # notify of derived duration property change
        self.connect("notify::start-time", self._notify_duration)
        self.connect("notify::end-time", self._notify_duration)

    def _notify_duration(self, *args):
        self.notify("duration")

    @GObject.Property(type=object)
    def duration(self) -> timedelta:
        return timedelta(microseconds=self.end_time.difference(self.start_time))

    def asdict(self) -> dict:
        return dict(
            start_time=self.start_time.format_iso8601(),
            end_time=self.end_time.format_iso8601(),
            task=self.task,
            client=self.client,
        )

    @classmethod
    def fromdict(cls, d: dict) -> "WorkedTime":
        tz = GLib.TimeZone.new_local()
        return cls(
            start_time=GLib.DateTime.new_from_iso8601(d["start_time"], tz),
            end_time=GLib.DateTime.new_from_iso8601(d["end_time"], tz),
            task=d["task"],
            client=d["client"],
        )

    def is_done(self) -> bool:
        return self.end_time is not None

    def is_started(self) -> bool:
        return self.start_time is not None

    def __str__(self):
        return f"WorkedTime({self.task}, {self.client}, {self.start_time}, {self.end_time})"


class WorkedTimeStore(Gio.ListStore):
    clients_changed = GObject.Signal("clients-changed")
    tasks_changed = GObject.Signal("tasks-changed")
    item_added = GObject.Signal("item-added", arg_types=(WorkedTime,))
    item_removed = GObject.Signal("item-removed", arg_types=(WorkedTime,))

    def __init__(self, filename: str):
        GObject.GObject.__init__(self)
        Gio.ListStore.__init__(self, item_type=WorkedTime)
        self._filename = filename
        self._clients = set()
        self._tasks = set()
        self.clients = Gtk.ListStore(str)
        self.tasks = Gtk.ListStore(str)

        # Load from file
        if os.path.exists(self._filename):
            with open(self._filename, "r") as f:
                for d in json.load(f):
                    self.append(WorkedTime.fromdict(d), suppress_signals=True)
        self._sort_by_start_time()  # just to be sure
        self._refresh_tasks()
        self._refresh_clients()

        # Only start saving when new items are added after the initial load
        self.connect("item-added", self.save)

    def _sort_by_start_time(self):
        def compare_start_times(a: WorkedTime, b: WorkedTime):
            # Note: GLib.DateTime.compare() is not available in Python apparently
            return a.start_time.to_unix() - b.start_time.to_unix()

        self.sort(compare_start_times)

    def get_subset(
            self,
            tasks: Optional[Set[str]] = None,
            clients: Optional[Set[str]] = None,
            start_time: Optional[GLib.DateTime] = None,
            end_time: Optional[GLib.DateTime] = None) -> Gio.ListStore:

        list_store = Gio.ListStore(item_type=WorkedTime)

        def is_in_subset(wt: WorkedTime):
            if start_time is not None and wt.start_time.to_unix() < start_time.to_unix():
                return False
            if end_time is not None and wt.start_time.to_unix() > end_time.to_unix():
                return False
            if tasks is not None and wt.task not in tasks:
                return False
            if clients is not None and wt.client not in clients:
                return False
            return True

        for wt in self:
            if is_in_subset(wt):
                list_store.append(wt)

        def _on_item_added(_wt_store, added_item):
            if is_in_subset(added_item):
                list_store.append(added_item)

        def _on_item_removed(_wt_store, removed_item):
            if is_in_subset(removed_item):
                found, position = list_store.find(removed_item)
                if found:
                    list_store.remove(position)
                else:
                    logging.error(f"Could not find item {removed_item} in subset")

        # TODO: if an item in the subset is changed, it should be removed if it no longer matches the subset
        # TODO: if an item outside the subset is changed, it should be added if it now matches the subset

        self.connect("item-added", _on_item_added)
        self.connect("item-removed", _on_item_removed)

        return list_store

    def save(self, *_args):
        logging.info(f"Saving worked time store to {self._filename}")
        with open(self._filename, "w") as f:
            json.dump([wt.asdict() for wt in self], f, indent=2)

    def insert(self, position: int, item: WorkedTime):
        super().insert(position, item)
        self._connect_item_to_signals(item)
        self.emit("item-added", item)

    def append(self, item: WorkedTime, suppress_signals: bool = False):
        super().append(item)
        self._connect_item_to_signals(item)
        if not suppress_signals:
            self.emit("item-added", item)

    def insert_sorted(self, item: WorkedTime, compare_func: Callable[[WorkedTime, WorkedTime], int], *user_data):
        raise NotImplementedError("Not implemented yet")

    def remove_item(self, item: WorkedTime):
        found, position = self.find(item)
        if found:
            self.remove(position)
        else:
            raise ValueError(f"Item not found: {item}")

    def remove(self, position: int):
        item = self[position]
        super().remove(position)
        self.emit("item-removed", item)

    def remove_all(self):
        for i in range(len(self)):
            self.remove(0)

    def _connect_item_to_signals(self, item: WorkedTime):
        item.connect("notify::client", self._refresh_clients)
        item.connect("notify::task", self._refresh_tasks)
        # Save to disk when item was changed
        item.connect("notify", self.save)

    def _refresh_clients(self, *_args):
        new_clients = {wt.client for wt in self}
        if new_clients != self._clients:
            added_clients = new_clients - self._clients
            removed_clients = self._clients - new_clients
            for client in added_clients:
                self.clients.append([client])
            for client in removed_clients:
                for row in self.clients:
                    if row[0] == client:
                        self.clients.remove(row.iter)

            self._clients = new_clients
            self.emit("clients-changed")

    def _refresh_tasks(self, *_args):
        new_tasks = {wt.task for wt in self}
        if new_tasks != self._tasks:
            added_tasks = new_tasks - self._tasks
            removed_tasks = self._tasks - new_tasks
            for task in added_tasks:
                self.tasks.append([task])
            for task in removed_tasks:
                for row in self.tasks:
                    if row[0] == task:
                        self.tasks.remove(row.iter)

            self._tasks = new_tasks
            self.emit("tasks-changed")

    def most_recent_worked_tasks_and_clients(self, n: int) -> Generator[Tuple[str, str], None, None]:
        """
        Yields the most recent n task-client-tuples.
        If a task-client-tuple is worked on multiple times, it is only yielded once.

        :param n: The number of task-client-tuples to yield (at most)
        :return: A generator yielding the most recent n task-client-tuples.
        """
        work_items = set()
        for worked_time in reversed(self):
            work_item = (worked_time.task, worked_time.client)
            if work_item not in work_items:
                work_items.add(work_item)
                yield work_item
            if len(work_items) >= n:
                break


import gi

from wage_labor_record.history_view.summary_view import SummaryView
from wage_labor_record.tracking_state import TrackingState

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

from wage_labor_record.worked_time_store import WorkedTimeStore
from wage_labor_record.history_view.selector_widget import SelectorWidget
from wage_labor_record.history_view.worked_times_list_view import WorkedTimesListView


class HistoryBrowserWindow(Gtk.Window):
    def __init__(self, tracking_state: TrackingState, work_time_store: WorkedTimeStore):
        super().__init__(title="Worked Time")
        self.set_default_size(200, 100)

        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        box.show()
        self.add(box)

        selector_box = SelectorWidget(work_time_store)
        selector_box.show()
        box.add(selector_box)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        scrolled_window.show()
        box.pack_start(scrolled_window, True, True, 0)  # Expand=True

        self.worked_time_widget = WorkedTimesListView(work_time_store)
        self.worked_time_widget.show()
        scrolled_window.add(self.worked_time_widget)

        self.summary_view = SummaryView(tracking_state)
        box.add(self.summary_view)

        def on_selection_changed(selector: SelectorWidget):
            subset = work_time_store.get_subset(
                tasks=selector.selected_tasks,
                clients=selector.selected_clients,
                start_time=selector.selected_start_time,
                end_time=selector.selected_end_time
            )
            self.worked_time_widget.set_worked_times_list(subset)
            self.summary_view.set_worked_times_list(
                subset,
                include_tracking_state=selector.selected_end_time is None)
        selector_box.connect("selection-changed", on_selection_changed)



import datetime

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib, GObject


class DatetimePicker(Gtk.Popover):

    datetime_changed = GObject.Signal("datetime-changed", arg_types=(GLib.DateTime,))

    def __init__(self, initial_time: GLib.DateTime):
        Gtk.Popover.__init__(self)
        self.set_position(Gtk.PositionType.BOTTOM)
        self.set_border_width(10)

        # create horizontal box
        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        self.add(box)

        # Create a date picker
        self.date_picker = Gtk.Calendar(
            day=initial_time.get_day_of_month(),
            month=initial_time.get_month()-1,
            year=initial_time.get_year(),
        )
        self.date_picker.connect("day-selected", self._emit_change_signal)
        box.add(self.date_picker)

        # Create a time picker with 10-minute increments
        self.hour_picker = Gtk.SpinButton()
        self.hour_picker.set_orientation(Gtk.Orientation.VERTICAL)
        self.hour_picker.set_range(-1, 24)
        self.hour_picker.set_increments(1, 1)
        self.hour_picker.set_numeric(True)
        self.hour_picker.set_value(initial_time.get_hour())
        self.hour_picker.connect("value-changed", self._on_hour_changed)
        box.add(self.hour_picker)

        box.add(Gtk.Label(label=":"))

        self.minute_picker = Gtk.SpinButton()
        self.minute_picker.set_orientation(Gtk.Orientation.VERTICAL)
        self.minute_picker.set_range(-1, 60)
        self.minute_picker.set_increments(10, 10)
        self.minute_picker.set_numeric(True)
        self.minute_picker.set_value(initial_time.get_minute())
        self.minute_picker.connect("value-changed", self._on_minute_changed)
        box.add(self.minute_picker)

    def _on_minute_changed(self, *_args):
        minute = self.minute_picker.get_value_as_int()
        if minute == 60:
            self.minute_picker.set_value(0)
            self.hour_picker.set_value(self.hour_picker.get_value_as_int() + 1)
        elif minute == -1:
            self.minute_picker.set_value(59)
            self.hour_picker.set_value(self.hour_picker.get_value_as_int() - 1)
        self._emit_change_signal()

    def _on_hour_changed(self, *_args):
        hour = self.hour_picker.get_value_as_int()
        if hour == 24:
            self.hour_picker.set_value(0)
            year, month, day = self.date_picker.get_date()
            next_day = datetime.date(year, month+1, day) + datetime.timedelta(days=1)
            self.date_picker.select_day(next_day.day)
            self.date_picker.select_month(next_day.month-1, next_day.year)
        elif hour == -1:
            self.hour_picker.set_value(23)
            year, month, day = self.date_picker.get_date()
            prev_day = datetime.date(year, month+1, day) - datetime.timedelta(days=1)
            self.date_picker.select_day(prev_day.day)
            self.date_picker.select_month(prev_day.month-1, prev_day.year)
        self._emit_change_signal()

    def _emit_change_signal(self, *_args):
        year, month, day = self.date_picker.get_date()
        hour = self.hour_picker.get_value_as_int()
        minute = self.minute_picker.get_value_as_int()
        self.emit("datetime-changed", GLib.DateTime.new_local(year, month + 1, day, hour, minute, 0))


from typing import Optional, Set

from gi.repository import GLib, GObject, Gtk, Gdk

from wage_labor_record.worked_time_store import WorkedTimeStore


class SelectorWidget(Gtk.Box):
    """Widget to select a subset of the history of tracked worked times."""
    selection_changed = GObject.Signal("selection-changed")

    def __init__(self, worked_time_store: WorkedTimeStore):
        Gtk.Box.__init__(self,
            orientation=Gtk.Orientation.VERTICAL,
            spacing=40,
            border_width=20,
        )

        self.selected_start_time: Optional[GLib.DateTime] = None
        self.selected_end_time: Optional[GLib.DateTime] = None
        self.selected_clients: Optional[Set[str]] = None
        self.selected_tasks: Optional[Set[str]] = None

        # Time Selector
        time_selections_model = Gtk.ListStore(str)
        time_selections_model.append(["Today"])
        time_selections_model.append(["This Week"])
        time_selections_model.append(["Last Week"])
        time_selections_model.append(["This Month"])
        time_selections_model.append(["Last Month"])
        time_selector = Gtk.TreeView()
        time_selector.set_model(time_selections_model)

        time_selector.append_column(
            Gtk.TreeViewColumn("Time Selection", Gtk.CellRendererText(), text=0))
        time_selector.show()

        def on_time_selector_changed(_):
            model, treeiter = time_selector.get_selection().get_selected()
            if treeiter is not None:
                time_selection = model[treeiter][0]
                if time_selection == "Today":
                    now = GLib.DateTime.new_now_local()
                    start_of_day = GLib.DateTime.new(now.get_timezone(), now.get_year(), now.get_month(), now.get_day_of_month(), 0, 0, 0)
                    self.selected_start_time = start_of_day
                    self.selected_end_time = None

                elif time_selection == "This Week":
                    now = GLib.DateTime.new_now_local()
                    start_of_today = GLib.DateTime.new(
                        now.get_timezone(),
                        now.get_year(),
                        now.get_month(),
                        now.get_day_of_month(),
                        0,
                        0,
                        0)
                    start_of_week = start_of_today.add_days(-now.get_day_of_week() + 1)
                    self.selected_start_time = start_of_week
                    self.selected_end_time = None

                elif time_selection == "Last Week":
                    now = GLib.DateTime.new_now_local()
                    start_of_today = GLib.DateTime.new(
                        now.get_timezone(),
                        now.get_year(),
                        now.get_month(),
                        now.get_day_of_month(),
                        0,
                        0,
                        0)
                    start_of_week = start_of_today.add_days(-now.get_day_of_week() + 1)
                    start_of_prev_week = start_of_week.add_weeks(-1)
                    self.selected_start_time = start_of_prev_week
                    self.selected_end_time = start_of_week

                elif time_selection == "This Month":
                    now = GLib.DateTime.new_now_local()
                    start_of_month = GLib.DateTime.new(now.get_timezone(), now.get_year(), now.get_month(), 1, 0, 0, 0)
                    self.selected_start_time = start_of_month
                    self.selected_end_time = None

                elif time_selection == "Last Month":
                    now = GLib.DateTime.new_now_local()
                    start_of_month = GLib.DateTime.new(now.get_timezone(), now.get_year(), now.get_month(), 1, 0, 0, 0)
                    start_of_prev_month = start_of_month.add_months(-1)
                    self.selected_start_time = start_of_prev_month
                    self.selected_end_time = start_of_month
                else:
                    assert False, f"Unknown time selection: {time_selection}"
                self.selection_changed.emit()

        time_selector.connect("cursor-changed", on_time_selector_changed)
        self.add(time_selector)

        # Task and Client Selectors (multiple selection)

        def on_esc_deselect_all(widget, event):
            if event.keyval == Gdk.KEY_Escape:
                widget.get_selection().unselect_all()

        task_selector = Gtk.TreeView(rubber_banding=True)
        task_selector.connect("key-press-event", on_esc_deselect_all)
        task_selector.set_model(worked_time_store.tasks)
        task_selector.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)
        task_selector.append_column(Gtk.TreeViewColumn("Task", Gtk.CellRendererText(), text=0))

        def on_task_selector_changed(_):
            if task_selector.get_selection().count_selected_rows() == 0:
                self.selected_tasks = None
            else:
                model, rows = task_selector.get_selection().get_selected_rows()
                self.selected_tasks = {model[row][0] for row in rows}
            self.selection_changed.emit()

        task_selector.get_selection().connect("changed", on_task_selector_changed)
        task_selector.show()

        self.add(task_selector)

        client_selector = Gtk.TreeView(rubber_banding=True)
        client_selector.connect("key-press-event", on_esc_deselect_all)
        client_selector.set_model(worked_time_store.clients)
        client_selector.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)
        client_selector.append_column(Gtk.TreeViewColumn("Client", Gtk.CellRendererText(), text=0))

        def on_client_selector_changed(_):
            if client_selector.get_selection().count_selected_rows() == 0:
                self.selected_clients = None
            else:
                model, rows = client_selector.get_selection().get_selected_rows()
                self.selected_clients = {model[row][0] for row in rows}
            self.selection_changed.emit()

        client_selector.get_selection().connect("changed", on_client_selector_changed)
        client_selector.show()

        self.add(client_selector)

        self.show()



import os
import sys
import warnings
from pathlib import Path
from typing import Optional, Set, Tuple

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import GLib, Gio, Gtk


def user_data_dir(app_name: str) -> Path:
    r"""
    Get OS specific data directory path for SwagLyrics.
    Typical user data directories are:
        macOS:    ~/Library/Application Support/<app_name>
        Unix:     ~/.local/share/<app_name>   # or in $XDG_DATA_HOME, if defined
        Win 10:   C:\Users\<username>\AppData\Local\<app_name>
    For Unix, we follow the XDG spec and support $XDG_DATA_HOME if defined.
    :return: full path to the user-specific data dir
    """

    # get os specific path
    if sys.platform.startswith("win"):
        os_path = os.getenv("LOCALAPPDATA")
    elif sys.platform.startswith("darwin"):
        os_path = "~/Library/Application Support"
    else:
        # linux
        os_path = os.getenv("XDG_DATA_HOME", "~/.local/share")

    # append app name
    path = Path(os_path) / app_name
    return path.expanduser()


def get_idle_time():
    # TODO: maybe this is a better way to do idle time detection:
    #  https://stackoverflow.com/questions/217157/how-can-i-determine-the-display-idle-time-from-python-in-windows-linux-and-mac
    if sys.platform=="win32":
        import ctypes
        return ctypes.windll.user32.GetLastInputInfo() / 1000
    elif sys.platform=="linux":
        import subprocess
        return int(subprocess.check_output(["xprintidle"])) / 1000
    else:
        warnings.warn(f"No idle time detection for platform {sys.platform}")
        return 0


def link_gtk_menu_item_to_gio_action(menu_item: Gtk.MenuItem, action: Gio.SimpleAction, parameter: Optional[GLib.Variant] = None):
    """
    Links a Gtk.MenuItem to a Gio.SimpleAction.
    This is an ugly hack needed because Gtk.MenuItem does not support Gio.SimpleAction.
    However, we can't use Gio.Menu because it does not support Xapp.StatusIcon.
    And we need Xapp.StatusIcon because it is supported by awesome-wm as opposed to Gtk.StatusIcon.
    """
    menu_item.connect("activate", lambda _0: action.activate(parameter))
    menu_item.set_sensitive(action.get_enabled())
    action.connect("notify::enabled", lambda _0, _1: menu_item.set_sensitive(action.get_enabled()))


def make_completer(model: Gtk.ListStore) -> Gtk.EntryCompletion:
    completer = Gtk.EntryCompletion(
        model=model,
        inline_completion=True,
        inline_selection=True,
        popup_completion=True,
        minimum_key_length=0,
    )
    # TODO: this is a hack, but it works
    completer.set_text_column(0)
    return completer

def filter_duplicate_items(model, iter, data: Tuple[int, set]) -> bool:
    """
    Used as a filter function for Gtk.TreeModelFilter.

    Use with
    >>> filter.set_visible_func(filter_duplicate_items, (column, set())
    """
    column, seen_items = data
    value = model[iter][column]
    if value in seen_items:
        return False
    seen_items.add(value)
    return True



import gi

from wage_labor_record.history_view.datetime_picker import DatetimePicker
from wage_labor_record.utils import make_completer

gi.require_version("Gtk", "3.0")
from gi.repository import GObject, Gio, Gtk

from wage_labor_record.worked_time_store import WorkedTime, WorkedTimeStore


class WorkedTimesListView(Gtk.ListBox):
    """A view to show and edit a set of worked times."""

    def __init__(self, worked_time_store: WorkedTimeStore):
        super().__init__()
        self.show()
        self._worked_times: Gio.ListStore = None
        self._all_items_in_same_year = False
        self._all_items_in_same_month = False
        self._all_items_in_same_day = False

        self._max_task_chars = 0
        self._max_client_chars = 0

        self._worked_time_store = worked_time_store

    def set_worked_times_list(self, model: Gio.ListStore):
        self._all_items_in_same_year = len({item.start_time.get_year() for item in model}) == 1
        self._all_items_in_same_month = self._all_items_in_same_year and len({item.start_time.get_month() for item in model}) == 1
        self._all_items_in_same_day = self._all_items_in_same_month and len({item.start_time.get_day_of_month() for item in model}) == 1

        self._max_task_chars = max(len(item.task) for item in model) if len(model) > 0 else 0
        self._max_client_chars = max(len(item.client) for item in model) if len(model) > 0 else 0

        self.bind_model(model, self._create_row)

    def _create_row(self, item: WorkedTime):
        row = Gtk.ListBoxRow()

        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        box.homogenous = False
        box.show()
        row.add(box)

        box.pack_start(self._create_task_entry(item), True, True, 0)
        box.pack_start(self._create_client_entry(item), True, True, 0)
        box.pack_start(self._create_start_time_button(item), False, False, 0)
        to_label = Gtk.Label("to", xalign=0)
        to_label.show()
        box.pack_start(to_label, False, False, 0)
        box.pack_start(self._create_end_time_button(item), False, False, 0)
        box.pack_start(self._create_delete_button(item), False, False, 0)  # don't expand the delete button
        box.show()
        row.show()
        return row

    def _create_start_time_button(self, item):
        start_time_button = Gtk.Button()
        start_time_button.set_relief(Gtk.ReliefStyle.NONE)

        def on_start_time_clicked(*args):
            picker = DatetimePicker(item.start_time)

            picker.connect("datetime-changed", lambda picker, time: item.set_property("start-time", time))
            picker.set_relative_to(start_time_button)
            picker.show_all()
            picker.popup()

        def on_start_time_changed(*args):
            start_time_button.set_label(self._get_start_time_string(item))

        on_start_time_changed()
        item.connect("notify::start-time", on_start_time_changed)
        start_time_button.connect("clicked", on_start_time_clicked)
        start_time_button.show()
        return start_time_button

    def _create_end_time_button(self, item):
        end_time_label = Gtk.Button()
        end_time_label.set_relief(Gtk.ReliefStyle.NONE)

        def on_end_time_clicked(*args):
            picker = DatetimePicker(item.end_time)

            picker.connect("datetime-changed", lambda picker, time: item.set_property("end-time", time))
            picker.set_relative_to(end_time_label)
            picker.show_all()
            picker.popup()

        def on_end_time_changed(*args):
            end_time_label.set_label(self._get_end_time_string(item))

        on_end_time_changed()
        item.connect("notify::end-time", on_end_time_changed)
        end_time_label.connect("clicked", on_end_time_clicked)
        end_time_label.show()
        return end_time_label

    def _create_client_entry(self, item: WorkedTime):
        client_entry = Gtk.Entry(
            text=item.client,
            placeholder_text="Client",
            completion=make_completer(self._worked_time_store.clients),
        )
        client_entry.set_width_chars(self._max_client_chars)
        client_entry.set_has_frame(False)

        item.bind_property("client", client_entry, "text", GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE)
        client_entry.show()
        return client_entry

    def _create_task_entry(self, item: WorkedTime):
        task_entry = Gtk.Entry(
            text=item.task,
            placeholder_text="Task",
            completion=make_completer(self._worked_time_store.tasks),
        )
        task_entry.set_width_chars(self._max_task_chars)
        task_entry.set_has_frame(False)

        item.bind_property("task", task_entry, "text", GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE)
        task_entry.show()
        return task_entry

    def _create_delete_button(self, item: WorkedTime):
        delete_button = Gtk.Button()
        delete_button.set_relief(Gtk.ReliefStyle.NONE)
        delete_button.set_image(Gtk.Image.new_from_icon_name("edit-delete-symbolic", Gtk.IconSize.BUTTON))
        delete_button.connect("clicked", lambda *args: self._worked_time_store.remove_item(item))
        delete_button.show()
        return delete_button

    def _get_start_time_string(self, item: WorkedTime) -> str:
        if self._all_items_in_same_day:
            return item.start_time.format("%H:%M")
        elif self._all_items_in_same_month:
            return item.start_time.format("%d %H:%M")
        elif self._all_items_in_same_year:
            return item.start_time.format("%b %d %H:%M")
        else:
            return item.start_time.format("%Y-%m-%d %H:%M")

    def _get_end_time_string(self, item: WorkedTime) -> str:
        ends_on_same_year = item.end_time.get_year() == item.start_time.get_year()
        ends_on_same_month = ends_on_same_year and item.end_time.get_month() == item.start_time.get_month()
        ends_on_same_day = ends_on_same_month and item.end_time.get_day_of_month() == item.start_time.get_day_of_month()

        if ends_on_same_day:
            return item.end_time.format("%H:%M")
        elif ends_on_same_month:
            return item.end_time.format("%d %H:%M")
        elif ends_on_same_year:
            return item.end_time.format("%b %d %H:%M")
        else:
            return item.end_time.format("%Y-%m-%d %H:%M")



import logging
import sys

import gi

from wage_labor_record.actions import AbortTrackingAction, SetCurrentTaskAction, StartTrackingAction, StopTrackingAction
from wage_labor_record.time_tracker_tray_icon import TimeTrackerTrayIcon
from wage_labor_record.time_tracker_window import TimeTrackerWindow
from wage_labor_record.worked_time_store import WorkedTimeStore
from wage_labor_record.tracking_state import TrackingState
from wage_labor_record.utils import get_idle_time, user_data_dir

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GLib

logging.basicConfig(level=logging.INFO)


class TimerTrackerApplication(Gtk.Application):

    def __init__(self):
        super().__init__(
            register_session=True,
            application_id="net.ernestum.wage_labor_record",
        )

        data_dir = user_data_dir("Wage Labor Record")
        data_dir.mkdir(parents=True, exist_ok=True)

        self.tracking_state = tracking_state = TrackingState(data_dir / "state.json")
        self.start_tracking_action = start_tracking_action = StartTrackingAction(tracking_state)
        self.start_tracking_task_action = start_tracking_task_action = SetCurrentTaskAction(tracking_state)
        self.stop_tracking_action = stop_tracking_action = StopTrackingAction(tracking_state)
        self.abort_tracking_action = abort_tracking_action = AbortTrackingAction(tracking_state)
        self.add_action(start_tracking_action)
        self.add_action(start_tracking_task_action)
        self.add_action(stop_tracking_action)
        self.add_action(abort_tracking_action)

        self.worked_time_store = worked_time_store = WorkedTimeStore(data_dir / "worked_times.json")

        stop_tracking_action.connect("worked-time", lambda _, worked_time: worked_time_store.append(worked_time))
        self.tray_icon = TimeTrackerTrayIcon(tracking_state, worked_time_store, self)

        def _check_for_idle(*_):
            if tracking_state.is_tracking():
                idle_time = get_idle_time()
                logging.debug(f"Idle time: {idle_time}")
                if idle_time > 60 * 15:  # when idle for 15 minutes
                    # show dialog to ask whether to continue tracking, stopping tracking and discarding the time or stopping tracking and saving the time
                    dialog = Gtk.MessageDialog(
                        transient_for=self.get_active_window(),
                        modal=True,
                        message_type=Gtk.MessageType.QUESTION,
                        buttons=Gtk.ButtonsType.NONE,
                        text="You have been idle for 15 minutes. Do you want to continue tracking?",
                    )
                    dialog.add_button("Continue", Gtk.ResponseType.YES)
                    dialog.add_button("Continue but discard (not implemented)", Gtk.ResponseType.APPLY).set_sensitive(False)
                    dialog.add_button("Stop and discard (not implemented)", Gtk.ResponseType.NO).set_sensitive(False)
                    dialog.add_button("Stop and save", Gtk.ResponseType.CANCEL).set_action_name("app.abort-tracking")
                    response = dialog.run()
                    # TODO: what do we do if the task has not been set yet? then we could not stop tracking
                    # TODO: add option to enter task in dialog
                    if response == Gtk.ResponseType.YES:
                        pass
                    elif response == Gtk.ResponseType.APPLY:
                        pass  # TODO: implement
                    elif response == Gtk.ResponseType.NO:
                        pass  # TODO: implement
                    elif response == Gtk.ResponseType.CANCEL:
                        stop_tracking_action.activate()

                    dialog.destroy()
            return True

        GLib.timeout_add(1000, _check_for_idle)

    def do_activate(self):
        self.hold()  # Keep the application running until we explicitly quit
        self.show_window()

    def show_window(self):
        if len(self.get_windows())==0:
            TimeTrackerWindow(
                tracking_state=self.tracking_state,
                worked_time_store=self.worked_time_store,
                application=self, title="Working Labor Record").present()

    def on_quit(self, action, param):
        self.worked_time_store.save()
        self.quit()

def main():
    app = TimerTrackerApplication()
    try:
        app.run(sys.argv)
    except KeyboardInterrupt:
        pass



import datetime

import gi

from wage_labor_record.tracking_state import TrackingState

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GLib


class SummaryView(Gtk.Box):
    """A widget to show summary information about a set of worked times."""
    def __init__(self, tracking_state: TrackingState):
        super().__init__(
            orientation=Gtk.Orientation.VERTICAL,
            spacing=40,
            border_width=20,
        )
        self._tracking_state = tracking_state

        self.total_time_label = Gtk.Label()
        self.total_time_label.set_markup(f"<span font='monospace bold 24'>00:00</span>")
        self.total_time_label.show()
        self.add(self.total_time_label)

        self.durations_by_task = Gtk.TreeView()

        self.durations_by_task.set_size_request(-1, 3 * 24)  # Ensure that the list is at least 3 lines tall
        self.durations_by_task.get_selection().set_mode(Gtk.SelectionMode.NONE)  # Disable selection
        self.durations_by_task.append_column(Gtk.TreeViewColumn("Task", Gtk.CellRendererText(), text=0))
        self.durations_by_task.append_column(Gtk.TreeViewColumn("Total Duration", Gtk.CellRendererText(), text=1))

        self.durations_by_task.show()
        self.add(self.durations_by_task)

        self._durations_by_task_string = ""
        self.copy_to_clipboard_button = Gtk.Button(label="Copy to Clipboard")
        def copy_to_clipboard(*args):
            clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
            clipboard.set_text(self._durations_by_task_string, -1)
        self.copy_to_clipboard_button.connect("clicked", copy_to_clipboard)
        self.copy_to_clipboard_button.show()
        self.add(self.copy_to_clipboard_button)
        self.show()

    def set_worked_times_list(self, worked_times_list, include_tracking_state: bool = False):
        # Compute the durations aggregated by task
        durations_by_task = dict()
        for worked_time in worked_times_list:
            durations_by_task.setdefault(worked_time.task, datetime.timedelta())
            durations_by_task[worked_time.task] += worked_time.duration

        # Save the durations by task as a string for copying to the clipboard
        self._durations_by_task_string = "\n".join([f'{task}, {_duration_to_str(duration)}' for task, duration in durations_by_task.items()])

        # Put the aggregated durations into a Gtk.ListStore and display it in the list view
        durations_by_task_list = Gtk.ListStore(str, str)
        for task, duration in durations_by_task.items():
            durations_by_task_list.append([task, _duration_to_str(duration)])
        self.durations_by_task.set_model(durations_by_task_list)

        # Compute the total duration
        total_duration = sum(durations_by_task.values(), start=datetime.timedelta())

        def update_total_duration_view():
            if include_tracking_state and self._tracking_state.is_tracking():
                total_duration_with_tracking_state = total_duration + self._tracking_state.elapsed_time()
                self.total_time_label.set_markup(
                    f"<span font='monospace bold 24'>{_duration_to_str(total_duration, include_seconds=False)}</span>\n"
                    f"<span font='monospace bold 16' color='grey'>({_duration_to_str(total_duration_with_tracking_state)})</span>"
                )
            else:
                self.total_time_label.set_markup(f"<span font='monospace bold 24'>{_duration_to_str(total_duration, include_seconds=False)}</span>")

        update_total_duration_view()
        if include_tracking_state:
            self._tracking_state.connect("notify", lambda *args: update_total_duration_view())

            # regularly update the total time label
            # When the tracking is active, repeatedly update the elapsed time label
            # when tracking stopped, stop the regular updates as well
            def _update_view():
                update_total_duration_view()
                return self._tracking_state.is_tracking()

            GLib.timeout_add(1000, _update_view)
            self._tracking_state.connect("notify::start-time", lambda *args: GLib.timeout_add(1000, _update_view))


def _duration_to_str(d: datetime.timedelta, include_seconds: bool = True) -> str:
    """Format the duration to HH:mm:ss format"""
    hours, remainder = divmod(int(d.total_seconds()), 60 * 60)
    minutes, seconds = divmod(remainder, 60)
    if include_seconds:
        return f"{hours:02}:{minutes:02}:{seconds:02}"
    else:
        return f"{hours:02}:{minutes:02}"


import importlib.resources
import logging

import gi

from wage_labor_record.tracking_state import TrackingState
from wage_labor_record.utils import link_gtk_menu_item_to_gio_action
from wage_labor_record.worked_time_store import WorkedTimeStore
from wage_labor_record.history_view.history_browser_window import HistoryBrowserWindow

gi.require_version("Gtk", "3.0")
gi.require_version('XApp', '1.0')
from gi.repository import GLib, Gtk, XApp


class TimeTrackerTrayIcon(XApp.StatusIcon):
    def __init__(self, tracking_state: TrackingState, worked_time_store: WorkedTimeStore, application: Gtk.Application):
        super().__init__()
        self.set_name("Time Tracker")

        def _update_tooltip(*_):
            task = tracking_state.task
            client = tracking_state.client
            is_tracking = tracking_state.is_tracking()

            tooltip_text = (
                    ("Working\n" if is_tracking else "Not working\n") +
                    (f"On: {task}\n" if task != "" else "") +
                    (f"For: {client}\n" if client != "" else "")
            ).strip()
            self.set_tooltip_text(tooltip_text)

        _update_tooltip()
        tracking_state.connect("notify::start-time", _update_tooltip)
        tracking_state.connect("notify::task", _update_tooltip)
        tracking_state.connect("notify::client", _update_tooltip)

        start_tracking_action = application.lookup_action("start_tracking")
        start_tracking_task_action = application.lookup_action("start_tracking_task")
        stop_tracking_action = application.lookup_action("stop_tracking")
        abort_tracking_action = application.lookup_action("abort_tracking")

        def on_left_click(*_):
            logging.debug("Left click")
            if start_tracking_action.get_enabled():
                start_tracking_action.activate()
            elif stop_tracking_action.get_enabled():
                stop_tracking_action.activate()
            else:
                application.activate()

        self.connect("activate", on_left_click)

        def _update_icon(*_):
            logging.debug("Update icon")
            if start_tracking_action.get_enabled():
                icon_name = "start-tracking"
            elif stop_tracking_action.get_enabled():
                icon_name = "stop-tracking"
            else:
                icon_name = "stop-tracking-disabled"

            with importlib.resources.path("wage_labor_record.resources", f"{icon_name}.svg") as p:
                self.set_icon_name(str(p))

        _update_icon()
        start_tracking_action.connect("notify::enabled", _update_icon)
        stop_tracking_action.connect("notify::enabled", _update_icon)

        # Add menu to the tray icon
        def _update_menu(*_):
            # NOTE: This is not the most efficient way to do this, but it works for now
            menu = Gtk.Menu()

            # CLIENT ----------------------------
            client_item = Gtk.MenuItem(label="")

            def _update_client_menu_item(*_):
                client = tracking_state.client
                client_item.set_label("Set Client" if client == "" else f"Client: {client}")

            _update_client_menu_item()
            tracking_state.connect("notify::client", _update_client_menu_item)
            client_item.connect("activate", lambda _0: application.activate())
            menu.append(client_item)

            # TASK ----------------------------
            task_item = Gtk.MenuItem(label="")

            def _update_task_menu_item(*_):
                task = tracking_state.task
                task_item.set_label("Set Task" if task=="" else f"Task: {task}")

            _update_task_menu_item()
            tracking_state.connect("notify::task", _update_task_menu_item)
            task_item.connect("activate", lambda _0: application.activate())
            menu.append(task_item)

            # SEPARATOR ----------------------------
            menu.append(Gtk.SeparatorMenuItem())

            # WORKED ITEMS -------------------------
            for task, client in worked_time_store.most_recent_worked_tasks_and_clients(5):

                worked_time_item = Gtk.MenuItem(label=f"\u25B6 {client} - {task}")
                worked_time_item.client = client
                worked_time_item.task = task

                link_gtk_menu_item_to_gio_action(worked_time_item, start_tracking_task_action, GLib.Variant("(ss)", (client, task)))
                menu.append(worked_time_item)

            # SEPARATOR ----------------------------
            menu.append(Gtk.SeparatorMenuItem())

            # HISTORY ----------------------------
            history_item = Gtk.MenuItem(label="History")
            history_item.connect("activate", lambda _0: HistoryBrowserWindow(tracking_state, worked_time_store).show_all())
            menu.append(history_item)

            # SEPARATOR ----------------------------
            menu.append(Gtk.SeparatorMenuItem())


            # ABORT ----------------------------
            abort_item = Gtk.MenuItem(label="Abort")
            link_gtk_menu_item_to_gio_action(abort_item, abort_tracking_action)
            menu.append(abort_item)

            # QUIT ----------------------------
            quit_item = Gtk.MenuItem(label="Quit")
            quit_item.connect("activate", lambda _0: application.quit())
            menu.append(quit_item)

            menu.show_all()
            self.set_secondary_menu(menu)

        _update_menu()

        worked_time_store.connect("items-changed", _update_menu)
